class MultipleSnapshotsException(Exception):
    pass


class MultipleImportSnapshotTasksException(Exception):
    pass


class MultipleImagesException(Exception):
    pass


class BucketDoesNotExistException(Exception):
    def __init__(self, bucket_name: str, *args, **kwargs):
        msg = f"The bucket named '{bucket_name}' does not exist. You will need to create the bucket before proceeding."
        super().__init__(msg, *args, **kwargs)



import glob
import os
import pathlib

import pytest
from pydantic import ValidationError
from ruamel.yaml.constructor import DuplicateKeyError

from awspub import context

curdir = pathlib.Path(__file__).parent.resolve()


def test_context_create():
    """
    Create a Context object from a given configuration
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    assert ctx.conf["source"]["path"] == curdir / "fixtures/config1.vmdk"
    assert ctx.source_sha256 == "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8"
    assert ctx.conf["source"]["architecture"] == "x86_64"
    assert ctx.conf["s3"]["bucket_name"] == "bucket1"


def test_context_create_minimal():
    """
    Create a Context object from a given minimal configuration
    """
    ctx = context.Context(curdir / "fixtures/config-minimal.yaml", None)
    assert ctx.conf["source"]["path"] == curdir / "fixtures/config1.vmdk"
    assert ctx.source_sha256 == "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8"
    assert ctx.conf["source"]["architecture"] == "x86_64"
    assert ctx.conf["s3"]["bucket_name"] == "bucket1"


def test_context_create_with_mapping():
    """
    Create a Context object from a given configuration
    """
    ctx = context.Context(curdir / "fixtures/config2.yaml", curdir / "fixtures/config2-mapping.yaml")
    assert ctx.conf["source"]["path"] == curdir / "fixtures/config1.vmdk"
    assert ctx.source_sha256 == "6252475408b9f9ee64452b611d706a078831a99b123db69d144d878a0488a0a8"
    assert ctx.conf["images"].get("test-image-value1")
    assert ctx.conf["images"].get("test-image-value$2")


def test_context_with_docs_config_samples():
    """
    Create a Context object with the sample config files used for documentation
    """
    config_samples_dir = curdir.parents[1] / "docs" / "config-samples"
    for f in glob.glob(f"{config_samples_dir}/*.yaml"):
        mapping_file = f + ".mapping"
        if os.path.exists(mapping_file):
            mapping = mapping_file
        else:
            mapping = None
        context.Context(os.path.join(config_samples_dir, f), mapping)


def test_context_with_duplicate_image_name():
    """
    Create a context with a configuration file that contains a duplicate image name key
    """
    with pytest.raises(DuplicateKeyError):
        context.Context(curdir / "fixtures/config3-duplicate-keys.yaml", None)


@pytest.mark.parametrize(
    "config_file",
    ["fixtures/config-minimal.yaml", "fixtures/config-valid-nonawspub.yaml"],
)
def test_valid_configuration(config_file):
    """
    Test with a valid configuration file (no extra fields)
    """
    ctx = context.Context(curdir / config_file, None)
    assert ctx.conf is not None
    assert ctx.conf["s3"]["bucket_name"] == "bucket1"
    assert ctx.conf["source"]["architecture"] == "x86_64"


def test_invalid_configuration_extra_field():
    """
    Test with an invalid configuration file that includes an extra field
    """
    with pytest.raises(ValidationError):
        context.Context(curdir / "fixtures/config-invalid-s3-extra.yaml", None)



import pathlib
from unittest.mock import MagicMock

import pytest

from awspub import context, snapshot

curdir = pathlib.Path(__file__).parent.resolve()


def test_snapshot__get_none_exist():
    """
    No snapshot exist - should return None
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    s = snapshot.Snapshot(ctx)
    client_mock = MagicMock()
    assert s._get(client_mock, "snapshot-name") is None
    client_mock.describe_snapshots.assert_called_with(
        Filters=[
            {"Name": "tag:Name", "Values": ["snapshot-name"]},
            {"Name": "status", "Values": ["pending", "completed"]},
        ],
        OwnerIds=["self"],
    )


def test_snapshot__get_one_exist():
    """
    One snapshot exist with the same name - should return the snapshot id
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    s = snapshot.Snapshot(ctx)
    client_mock = MagicMock()
    client_mock.describe_snapshots = MagicMock(return_value={"Snapshots": [{"SnapshotId": "snap-1"}]})
    assert s._get(client_mock, "snapshot-name") == "snap-1"
    client_mock.describe_snapshots.assert_called_with(
        Filters=[
            {"Name": "tag:Name", "Values": ["snapshot-name"]},
            {"Name": "status", "Values": ["pending", "completed"]},
        ],
        OwnerIds=["self"],
    )


def test_snapshot__get_multiple_exist():
    """
    Multiple snapshots exist - _get() should raise an Exception
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    s = snapshot.Snapshot(ctx)
    client_mock = MagicMock()
    client_mock.describe_snapshots = MagicMock(
        return_value={"Snapshots": [{"SnapshotId": "snap-1"}, {"SnapshotId": "snap-2"}]}
    )
    with pytest.raises(Exception):
        s._get(client_mock, "snapshot-name")
    client_mock.describe_snapshots.assert_called_with(
        Filters=[
            {"Name": "tag:Name", "Values": ["snapshot-name"]},
            {"Name": "status", "Values": ["pending", "completed"]},
        ],
        OwnerIds=["self"],
    )


def test_snapshot__get_import_snapshot_task_completed():
    """
    Test the Snapshot._get_import_snapshot_task() method
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    s = snapshot.Snapshot(ctx)
    client_mock = MagicMock()
    client_mock.describe_import_snapshot_tasks = MagicMock(
        return_value={
            "ImportSnapshotTasks": [
                {
                    "ImportTaskId": "import-snap-08b79d7b5d382d56b",
                    "SnapshotTaskDetail": {
                        "SnapshotId": "snap-0e0f3407a1b541c40",
                        "Status": "completed",
                    },
                    "Tags": [
                        {"Key": "Name", "Value": "021abb3f2338b5e57b5d870816565429659bc70769d71c486234ad60fe6aec67"},
                    ],
                }
            ],
        }
    )
    assert (
        s._get_import_snapshot_task(client_mock, "021abb3f2338b5e57b5d870816565429659bc70769d71c486234ad60fe6aec67")
        is None
    )


def test_snapshot__get_import_snapshot_task_active():
    """
    Test the Snapshot._get_import_snapshot_task() method
    """
    ctx = context.Context(curdir / "fixtures/config1.yaml", None)
    s = snapshot.Snapshot(ctx)
    client_mock = MagicMock()
    client_mock.describe_import_snapshot_tasks = MagicMock(
        return_value={
            "ImportSnapshotTasks": [
                {
                    "ImportTaskId": "import-snap-08b79d7b5d382d56b",
                    "SnapshotTaskDetail": {
                        "SnapshotId": "snap-0e0f3407a1b541c40",
                        "Status": "active",
                    },
                    "Tags": [
                        {"Key": "Name", "Value": "021abb3f2338b5e57b5d870816565429659bc70769d71c486234ad60fe6aec67"},
                    ],
                }
            ],
        }
    )
    assert (
        s._get_import_snapshot_task(client_mock, "021abb3f2338b5e57b5d870816565429659bc70769d71c486234ad60fe6aec67")
        == "import-snap-08b79d7b5d382d56b"
    )



#!/usr/bin/python3

import argparse
import json
import logging
import pathlib
import sys

import awspub

logger = logging.getLogger(__name__)


def _create(args) -> None:
    """
    Create images based on the given configuration and write json
    data to the given output
    """
    images_by_name, images_by_group = awspub.create(args.config, args.config_mapping, args.group)
    images_json = json.dumps({"images": images_by_name, "images-by-group": images_by_group}, indent=4)
    args.output.write(images_json)


def _list(args) -> None:
    """
    List images based on the given configuration and write json
    data to the given output
    """
    images_by_name, images_by_group = awspub.list(args.config, args.config_mapping, args.group)
    images_json = json.dumps({"images": images_by_name, "images-by-group": images_by_group}, indent=4)
    args.output.write(images_json)


def _verify(args) -> None:
    """
    Verify available images against configuration
    """
    problems = awspub.verify(args.config, args.config_mapping, args.group)
    args.output.write((json.dumps({"problems": problems}, indent=4)))


def _cleanup(args) -> None:
    """
    Cleanup available images
    """
    awspub.cleanup(args.config, args.config_mapping, args.group)


def _publish(args) -> None:
    """
    Make available images public
    """
    awspub.publish(args.config, args.config_mapping, args.group)


def _parser():
    parser = argparse.ArgumentParser(description="AWS EC2 publication tool")
    parser.add_argument("--log-level", choices=["info", "debug"], default="info")
    parser.add_argument("--log-file", type=pathlib.Path, help="write log to given file instead of stdout")
    parser.add_argument("--log-console", action=argparse.BooleanOptionalAction, help="write log to stdout")
    p_sub = parser.add_subparsers(help="sub-command help")

    # create
    p_create = p_sub.add_parser("create", help="Create images")
    p_create.add_argument(
        "--output", type=argparse.FileType("w+"), help="output file path. defaults to stdout", default=sys.stdout
    )
    p_create.add_argument("--config-mapping", type=pathlib.Path, help="the image config template mapping file path")
    p_create.add_argument("--group", type=str, help="only handles images from given group")
    p_create.add_argument("config", type=pathlib.Path, help="the image configuration file path")
    p_create.set_defaults(func=_create)

    # list
    p_list = p_sub.add_parser("list", help="List images (but don't modify anything)")
    p_list.add_argument(
        "--output", type=argparse.FileType("w+"), help="output file path. defaults to stdout", default=sys.stdout
    )
    p_list.add_argument("--config-mapping", type=pathlib.Path, help="the image config template mapping file path")
    p_list.add_argument("--group", type=str, help="only handles images from given group")
    p_list.add_argument("config", type=pathlib.Path, help="the image configuration file path")
    p_list.set_defaults(func=_list)

    # verify
    p_verify = p_sub.add_parser("verify", help="Verify images")
    p_verify.add_argument(
        "--output", type=argparse.FileType("w+"), help="output file path. defaults to stdout", default=sys.stdout
    )
    p_verify.add_argument("--config-mapping", type=pathlib.Path, help="the image config template mapping file path")
    p_verify.add_argument("--group", type=str, help="only handles images from given group")
    p_verify.add_argument("config", type=pathlib.Path, help="the image configuration file path")

    p_verify.set_defaults(func=_verify)

    # cleanup
    p_cleanup = p_sub.add_parser("cleanup", help="Cleanup images")
    p_cleanup.add_argument(
        "--output", type=argparse.FileType("w+"), help="output file path. defaults to stdout", default=sys.stdout
    )
    p_cleanup.add_argument("--config-mapping", type=pathlib.Path, help="the image config template mapping file path")
    p_cleanup.add_argument("--group", type=str, help="only handles images from given group")
    p_cleanup.add_argument("config", type=pathlib.Path, help="the image configuration file path")

    p_cleanup.set_defaults(func=_cleanup)

    # publish
    p_publish = p_sub.add_parser("publish", help="Publish images")
    p_publish.add_argument(
        "--output", type=argparse.FileType("w+"), help="output file path. defaults to stdout", default=sys.stdout
    )
    p_publish.add_argument("--config-mapping", type=pathlib.Path, help="the image config template mapping file path")
    p_publish.add_argument("--group", type=str, help="only handles images from given group")
    p_publish.add_argument("config", type=pathlib.Path, help="the image configuration file path")

    p_publish.set_defaults(func=_publish)

    return parser


def main():
    parser = _parser()
    args = parser.parse_args()
    log_formatter = logging.Formatter("%(asctime)s:%(name)s:%(levelname)s:%(message)s")
    # log level
    loglevel = logging.INFO
    if args.log_level == "debug":
        loglevel = logging.DEBUG
    root_logger = logging.getLogger()
    root_logger.setLevel(loglevel)
    # log file
    if args.log_file:
        file_handler = logging.FileHandler(filename=args.log_file)
        file_handler.setFormatter(log_formatter)
        root_logger.addHandler(file_handler)
    # log console
    if args.log_console:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(log_formatter)
        root_logger.addHandler(console_handler)
    if "func" not in args:
        sys.exit(parser.print_help())
    args.func(args)
    sys.exit(0)


if __name__ == "__main__":
    main()



import base64
import hashlib
import logging
import os
from typing import Dict

import boto3
from mypy_boto3_s3.type_defs import CompletedPartTypeDef

from awspub.context import Context
from awspub.exceptions import BucketDoesNotExistException

# chunk size is required for calculating the checksums
MULTIPART_CHUNK_SIZE = 8 * 1024 * 1024


logger = logging.getLogger(__name__)


class S3:
    """
    Handle S3 API interaction
    """

    def __init__(self, context: Context):
        """
        :param context:
        "type context: awspub.context.Context
        """
        self._ctx: Context = context
        self._s3client = boto3.client("s3")
        self._bucket_region = None

    @property
    def bucket_region(self):
        if not self._bucket_region:
            if not self._bucket_exists():
                raise BucketDoesNotExistException(self.bucket_name)
            self._bucket_region = self._s3client.head_bucket(Bucket=self.bucket_name)["BucketRegion"]

        return self._bucket_region

    @property
    def bucket_name(self):
        return self._ctx.conf["s3"]["bucket_name"]

    def __repr__(self):
        return (
            f"<{self.__class__} bucket:'{self.bucket_name}' "
            f"region:'{self.bucket_region} key:{self._ctx.source_sha256}'>"
        )

    def _multipart_sha256sum(self, file_path: str) -> str:
        """
        Calculate the sha256 checksum like AWS does it (in a multipart upload) per chunk
        See https://docs.aws.amazon.com/AmazonS3/latest/userguide/checking-object-integrity.html#large-object-checksums

        :param file_path: the path to the local file to upload
        :type file_path: str
        """
        sha256_list = []
        count = 0
        with open(file_path, "rb") as f:
            for chunk in iter(lambda: f.read(MULTIPART_CHUNK_SIZE), b""):
                sha256_list.append(hashlib.sha256(chunk))
                count += 1

        sha256_list_digest_concatenated = b"".join([s.digest() for s in sha256_list])
        sha256_b64 = base64.b64encode(hashlib.sha256(sha256_list_digest_concatenated).digest())
        return f"{sha256_b64.decode('ascii')}-{count}"

    def _bucket_exists(self) -> bool:
        """
        Check if the S3 bucket from context exists

        :return: True if the bucket exists, otherwise False
        :rtype: bool
        """
        resp = self._s3client.list_buckets()
        return self.bucket_name in [b["Name"] for b in resp["Buckets"]]

    def upload_file(self, source_path: str):
        """
        Upload a given file to the bucket from context. The key name will be the sha256sum hexdigest of the file.
        If a file with that name already exist in the given bucket and the calculated sha256sum matches
        the sha256sum from S3, nothing will be uploaded. Instead the existing file will be used.
        This method does use a multipart upload internally so an upload can be retriggered in case
        of errors and the previously uploaded content will be reused.
        Note: be aware that failed multipart uploads are not deleted. So it's recommended to setup
        a bucket lifecycle rule to delete incomplete multipart uploads.
        See https://docs.aws.amazon.com/AmazonS3/latest/userguide//mpu-abort-incomplete-mpu-lifecycle-config.html

        :param source_path: the path to the local file to upload (usually a .vmdk file)
        :type source_path: str
        """
        # make sure the bucket exists
        if not self._bucket_exists():
            raise BucketDoesNotExistException(self.bucket_name)

        s3_sha256sum = self._multipart_sha256sum(source_path)

        try:
            # check if the key exists already in the bucket and if so, if the multipart upload
            # sha256sum does match
            head = self._s3client.head_object(
                Bucket=self.bucket_name, Key=self._ctx.source_sha256, ChecksumMode="ENABLED"
            )

            if head["ChecksumSHA256"] == s3_sha256sum:
                logger.info(
                    f"'{self._ctx.source_sha256}' in bucket '{self.bucket_name}' "
                    "already exists and sha256sum matches. nothing to upload to S3"
                )
                return
            else:
                logger.warn(
                    f"'{self._ctx.source_sha256}' in bucket '{self.bucket_name}' "
                    f"already exists but sha256sum does not match. Will be overwritten ..."
                )
        except Exception:
            logging.debug(f"Can not find '{self._ctx.source_sha256}' in bucket '{self.bucket_name}'")

        # do the real upload
        self._upload_file_multipart(source_path, s3_sha256sum)

    def _get_multipart_upload_id(self) -> str:
        """
        Get an existing or create a multipart upload id

        :return: a multipart upload id
        :rtype: str
        """
        resp = self._s3client.list_multipart_uploads(Bucket=self.bucket_name)
        multipart_uploads = [
            upload["UploadId"] for upload in resp.get("Uploads", []) if upload["Key"] == self._ctx.source_sha256
        ]
        if len(multipart_uploads) == 1:
            logger.info(f"found existing multipart upload '{multipart_uploads[0]}' for key '{self._ctx.source_sha256}'")
            return multipart_uploads[0]
        elif len(multipart_uploads) == 0:
            # create a new multipart upload
            resp_create = self._s3client.create_multipart_upload(
                Bucket=self.bucket_name,
                Key=self._ctx.source_sha256,
                ChecksumAlgorithm="SHA256",
                ACL="private",
            )
            upload_id = resp_create["UploadId"]
            logger.info(
                f"new multipart upload (upload id: '{upload_id})' started in bucket "
                f"{self.bucket_name} for key {self._ctx.source_sha256}"
            )
            # if there's an expire rule configured for that bucket, inform about it
            if resp_create.get("AbortDate"):
                logger.info(
                    f"multipart upload '{upload_id}' will expire at "
                    f"{resp_create['AbortDate']} (rule: {resp_create.get('AbortRuleId')})"
                )
            else:
                logger.warning("there is no matching expire/lifecycle rule configured for incomplete multipart uploads")
            return upload_id
        else:
            # multiple multipart uploads for the same key available
            logger.warning(
                f"there are multiple ({len(multipart_uploads)}) multipart uploads ongoing in "
                f"bucket {self.bucket_name} for key {self._ctx.source_sha256}"
            )
            logger.warning("using the first found multipart upload but you should delete pending multipart uploads")
            return multipart_uploads[0]

    def _upload_file_multipart(self, source_path: str, s3_sha256sum: str) -> None:
        """
        Upload a given file to the bucket from context. The key name will be the sha256sum hexdigest of the file

        :param source_path: the path to the local file to upload (usually a .vmdk file)
        :type source_path: str
        :param s3_sha256sum: the sha256sum how S3 calculates it
        :type s3_sha256sum: str
        """
        upload_id = self._get_multipart_upload_id()

        logger.info(f"using upload id '{upload_id}' for multipart upload of '{source_path}' ...")
        resp_list_parts = self._s3client.list_parts(
            Bucket=self.bucket_name, Key=self._ctx.source_sha256, UploadId=upload_id
        )

        # sanity check for the used checksum algorithm
        if resp_list_parts["ChecksumAlgorithm"] != "SHA256":
            logger.error(f"available ongoing multipart upload '{upload_id}' does not use SHA256 as checksum algorithm")

        # already available parts
        parts_available = {p["PartNumber"]: p for p in resp_list_parts.get("Parts", [])}
        # keep a list of parts (either already available or created) required to complete the multipart upload
        parts: Dict[int, CompletedPartTypeDef] = {}
        parts_size_done: int = 0
        source_path_size: int = os.path.getsize(source_path)
        with open(source_path, "rb") as f:
            # parts start at 1 (not 0)
            for part_number, chunk in enumerate(iter(lambda: f.read(MULTIPART_CHUNK_SIZE), b""), start=1):
                # the sha256sum of the current part
                sha256_part = base64.b64encode(hashlib.sha256(chunk).digest()).decode("ascii")
                # do nothing if that part number already exist and the sha256sum matches
                if parts_available.get(part_number):
                    if parts_available[part_number]["ChecksumSHA256"] == sha256_part:
                        logger.info(f"part {part_number} already exists and sha256sum matches. continue")
                        parts[part_number] = dict(
                            PartNumber=part_number,
                            ETag=parts_available[part_number]["ETag"],
                            ChecksumSHA256=parts_available[part_number]["ChecksumSHA256"],
                        )
                        parts_size_done += len(chunk)
                        continue
                    else:
                        logger.info(f"part {part_number} already exists but will be overwritten")

                # upload a new part
                resp_upload_part = self._s3client.upload_part(
                    Body=chunk,
                    Bucket=self.bucket_name,
                    ContentLength=len(chunk),
                    ChecksumAlgorithm="SHA256",
                    ChecksumSHA256=sha256_part,
                    Key=self._ctx.source_sha256,
                    PartNumber=part_number,
                    UploadId=upload_id,
                )
                parts_size_done += len(chunk)
                # add new part to the dict of parts
                parts[part_number] = dict(
                    PartNumber=part_number,
                    ETag=resp_upload_part["ETag"],
                    ChecksumSHA256=sha256_part,
                )
                logger.info(
                    f"part {part_number} uploaded ({round(parts_size_done/source_path_size * 100, 2)}% "
                    f"; {parts_size_done} / {source_path_size} bytes)"
                )

        logger.info(
            f"finishing the multipart upload for key '{self._ctx.source_sha256}' in bucket {self.bucket_name} now ..."
        )
        # finish the multipart upload
        self._s3client.complete_multipart_upload(
            Bucket=self.bucket_name,
            Key=self._ctx.source_sha256,
            UploadId=upload_id,
            ChecksumSHA256=s3_sha256sum,
            MultipartUpload={"Parts": [value for key, value in parts.items()]},
        )
        logger.info(
            f"multipart upload finished and key '{self._ctx.source_sha256}' now "
            f"available in bucket '{self.bucket_name}'"
        )

        # add tagging to the final s3 object
        self._s3client.put_object_tagging(
            Bucket=self.bucket_name,
            Key=self._ctx.source_sha256,
            Tagging={
                "TagSet": self._ctx.tags,
            },
        )



import hashlib
import logging
import pathlib
from string import Template
from typing import Dict

from ruamel.yaml import YAML

from awspub.configmodels import ConfigModel

logger = logging.getLogger(__name__)


class Context:
    """
    Context holds the used configuration and some
    automatically calculated values
    """

    def __init__(self, conf_path: pathlib.Path, conf_template_mapping_path: pathlib.Path):
        self._conf_path = conf_path
        self._conf = None
        self._conf_template_mapping_path = conf_template_mapping_path
        self._conf_template_mapping = {}
        yaml = YAML(typ="safe")

        # read the config mapping first
        if self._conf_template_mapping_path:
            with open(self._conf_template_mapping_path, "r") as ctm:
                self._conf_template_mapping = yaml.load(ctm)
                logger.debug(f"loaded config template mapping for substitution: {self._conf_template_mapping}")

        # read the config itself
        with open(self._conf_path, "r") as f:
            template = Template(f.read())
            # substitute the values in the config with values from the config template mapping
            ft = template.substitute(**self._conf_template_mapping)
            y = yaml.load(ft)["awspub"]
            self._conf = ConfigModel(**y).model_dump()
            logger.debug(f"config loaded and validated as: {self._conf}")

        # handle relative paths in config files. those are relative to the config file dirname
        if not self.conf["source"]["path"].is_absolute():
            self.conf["source"]["path"] = pathlib.Path(self._conf_path).parent / self.conf["source"]["path"]

        for image_name, props in self.conf["images"].items():
            if props["uefi_data"] and not self.conf["images"][image_name]["uefi_data"].is_absolute():
                self.conf["images"][image_name]["uefi_data"] = (
                    pathlib.Path(self._conf_path).parent / self.conf["images"][image_name]["uefi_data"]
                )

        # calculate the sha256 sum of the source file once
        self._source_sha256_obj = self._sha256sum(self.conf["source"]["path"])
        self._source_sha256 = self._source_sha256_obj.hexdigest()

    @property
    def conf(self):
        return self._conf

    @property
    def source_sha256(self):
        """
        The sha256 sum hexdigest of the source->path value from the given
        configuration. This value is used in different places (eg. to automatically
        upload to S3 with this value as key)
        """
        return self._source_sha256

    @property
    def tags_dict(self) -> Dict[str, str]:
        """
        Common tags which will be used for all AWS resources
        This includes tags defined in the configuration file
        but doesn't include image group specific tags.
        Usually the tags() method should be used.
        """
        tags = dict()
        tags["awspub:source:filename"] = self.conf["source"]["path"].name
        tags["awspub:source:architecture"] = self.conf["source"]["architecture"]
        tags["awspub:source:sha256"] = self.source_sha256
        tags.update(self.conf.get("tags", {}))
        return tags

    @property
    def tags(self):
        """
        Helper to make tags directly usable by the AWS EC2 API
        which requires a list of dicts with "Key" and "Value" defined.
        """
        tags = []
        for name, value in self.tags_dict.items():
            tags.append({"Key": name, "Value": value})
        return tags

    def _sha256sum(self, file_path: pathlib.Path):
        """
        Calculate a sha256 sum for a given file

        :param file_path: the path to the local file to upload
        :type file_path: pathlib.Path
        :return: a haslib Hash object
        :rtype: _hashlib.HASH
        """
        sha256_hash = hashlib.sha256()
        with open(file_path.resolve(), "rb") as f:
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)
        return sha256_hash
